/* global Vue IMatchModals moment */
import '/system/vue/vue.min.js'

import fileWindow from '/system/ptc-vue/fw/ptc-filewindow.js'
import FileWindowLayout from '/system/ptc-vue/fw/ptc-fw-layout.js'
import FileWindowLoader from '/system/ptc-vue/fw/ptc-fw-loader.js'

import IMATEngine from './imat-engine.js'
import IMATServiceGoogle from './service-google.js'
import IMATServiceMicrosoft from './service-microsoft.js'
import IMATServiceClarifai from './service-clarifai.js'
import IMATServiceImagga from './service-imagga.js'
import IMATServiceIMatch from './imatch/service-imatch.js'

import IMATConfigure from './components/configure.js'

let DATA_CONFIG = 'photools.com.imatch.auto-tagger.{u}.config';
const DATA_CONFIG_VERSION = 2;
const LOG_NAME = 'AutoTagger '

$(document).ready(function () {

    let vm = new Vue({
        el: '#app',
        components: {
            'ptc-filewindow': fileWindow
        },
        data: {
            appConfig: {
                activeTab: 'service',

                expertMode: false,

                // If this is true, changing the focus to a different
                // file also updates the focus in the File Window.
                updateFileWindowFocus: true,

                // This size * 20 is the size used for the strip.
                thumbSize: 5,

                // Use square thumbnails?
                squareThumbs: false,

                // Show the mapped keywords (else the original keywords are shown)
                showMapped: false,

                // Sort keywords alphabetically
                sortAlphabetically: false,

                // The name of the selected service (from services.json)
                service: '',

                // Settings for each service in use
                serviceSettings: {},

                // Stats for each service in use
                serviceStats: {},

                // Show the keyboard hint on the lookup page?
                showKeyboardHints: true,

                processing: {
                    // '' = all services
                    '' : {
                        // if true, lookup processed keywords in the thesaurus
                        thesaurusLookup: {
                            enabled: true,
                            searchSynonyms: true,
                            includeSynonyms: true
                        },

                        // How to transform case: {none,'toLower','toUpper','toCamelCase'}
                        caseTransform: 'toLower',

                        // Use locale-specific lower/upper case functions.
                        // See: https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/toLocaleLowerCase
                        // .toLocaleLowerCase('en-US')
                        caseTransformUseLocale: true,

                        // If this is true, tests in the ignore list / mapping are performed
                        // case-sensitive.
                        caseSensitiveCompare: false,

                        // How many keywords to retrieve
                        maxKeywords: 20,

                        // Delete existing keywords of the file before assigning?
                        deleteExistingKeywords: false,

                        // Don't update files which already have keywords.
                        skipFilesWithKeywords: false,

                        // Don't assign unmapped keywords.
                        skipUnmapped: false,

                        // Assign processed files to this category
                        category: '',

                        // Empty category before assigning files
                        emptyCategory: false,

                        // If landmarks are enabled and GPS coordinates 
                        // are delivered, so we assign these coordinates
                        // to the image?
                        applyGPSCoordinates: false,

                        // Keywords to ignore
                        ignore: [],

                        // Mapping from source keyword to one or more target keywords (translation/synonyms)
                        mapping: {
                            // Do we map to a default parent keyword if no mapping exists?
                            defaultMappingEnabled: false,
                            // The parent keyword to map to: "beach" => "AUTO-TAGGER|beach"
                            defaultMapping: 'AUTO-TAGGER',
                            // Delivered keyword => array of mapped keywords.
                            maps: {}
                        }
                    }
                }
            },

            busy: false,
            firstLookup: true,

            runState: {
                running: false,
                canceled: false,
                percentage: 0,
                completed: false,
                filesProcessed: 0,
                filesSkipped: 0,
                keywordsAdded: 0,
                error: false,
                totalTime: 0
            },

            // For testing purposes
            simulate: false,

            databaseReadOnly: false,

            engine: null,

            // The contents of services.json
            serviceInfo: null,

            // The AI service to use
            serviceInstance: null,

            // To manage the file window strip
            fileWindowLayout: new FileWindowLayout(),
            fileWindowLoader: new FileWindowLoader(),

            // The files we are processing
            files: [],
            filesWithKeywords: [],
            selectedFiles: [],
            protectedFiles: [],

            // The currently focused file.
            focusedFile: 0,

            // This is true when we have not yet loaded a file.
            firstLoadedFile: true,
            
            // The data we fill for each file (original data and AI results)
            fileData: {},

            // The id of the assignment category for the current processing.
            // 0 if none.
            categoryId: 0,

            showProtectedFiles: false,

            // The new keyword to add
            newKeyword: '',

            stage: 'startPage'
        },

        beforeCreate() {
            // Has the user disabled protected files?
            IMatch.showProtectedFiles().then(settings => {
                this.showProtectedFiles = settings['protectedFilesVisible'].value;
            });

            IMatch.IMatchInfo().then(response => {
                this.IMatchInfo = response;
                this.databaseReadOnly = !response.database || response.databaseReadOnly;
                IMatchTranslate.init(IMatchTranslate.getUserLanguage(),'en','res/').then(() => {
                    IMatch.userInfo().then(response => {
                        DATA_CONFIG = DATA_CONFIG.replace('{u}',response.userName);
                        this.loadConfig();
                    });
                });
            });
        },

        mounted() {
            this.$el.style.display = 'block';
        },

        updated() {
            if (IMatchTranslate.isValid()) {
                IMatchTranslate.translateDOM();
            }
        },

        computed: {
            isDeveloperMode()
            {
                return this.IMatchInfo.developerMode
            },

            modelName () {
                let res = '';
                if (this.serviceInstance) {
                    const sinfo = this.serviceInfo[this.appConfig.service];
                    const ssettings = this.appConfig.serviceSettings[this.appConfig.service];
                    if (sinfo && ssettings) {
                        const model = sinfo.models[ssettings.model];
                        if (model) {
                            res = model.name[IMatchTranslate.getLoadedLangId()] || model.name['en'] || '';
                        }
                    }
                }
                return res;
            },

            serviceName()
            {
                if (this.serviceInstance) {
                    let url = '#';
                    if (this.serviceInfo[this.appConfig.service].infoUrl) {
                        url = this.serviceInfo[this.appConfig.service].infoUrl;
                    }
                    return IMatchTranslate.getStr('T_SERVICE_NAME').
                        replace('{0}',`<a href="#" @click="openLink('${url}')">${this.serviceInstance.name()}</a>`).
                        replace('{1}',this.modelName)
                }
                else {
                    return '';
                }
            },

            serviceLogo()
            {
                if (this.serviceInfo && (this.appConfig.service != '')) {
                    if (this.serviceInfo[this.appConfig.service].logo) {
                        return this.serviceInfo[this.appConfig.service].logo;
                    }
                    else {
                        return null;
                    }
                }
                else {
                    return null;
                }
            },

            // Returns true if the selected service supports landmarks.
            serviceSupportsLandmarks()
            {
                if (this.serviceInfo) {
                    let svc = this.serviceInfo[this.appConfig.service];
                    if (svc && svc.features && svc.features.landmarks) {
                        return true;
                    }
                }
                return false;
            },

            serviceInitialized() {
                return this.serviceInstance ? this.serviceInstance.initializationComplete() : false;
            },

            formatServiceInfo()
            {
                if (this.appConfig.service != '') {
                    let stats = this.appConfig.serviceStats[this.appConfig.service];
                    return IMatchTranslate.getStr('T_SERVICE_STATS').replace('{0}',stats.apiCalls).replace('{1}',moment.unix(stats.sinceDate).format('L LTS'));
                }
                else {
                    return '';
                }
            },

            formatServices()
            {
                let services = '';
                for (let i in this.serviceInfo) {
                    let s = this.serviceInfo[i];
                    if (services.length) services += ', ';
                    services += `<li><a href="${s.infoUrl}">${s.name['en']}</a></li>`;
                }
                return '<ul>' + IMatchTranslate.getStr('T_WELCOME_SERVICES').replace('{0}',services) + '</ul>';
            },

            getCurrentFileResults()
            {
                if (this.focusedFile && this.fileData[this.focusedFile]) {
                    return this.fileData[this.focusedFile].result;
                }
                else {
                    return null;
                }
            },

            hasKeywords()
            {
                const r = this.getCurrentFileResults;
                if (r) {
                    return r.keywords && (r.keywords.length > 0);
                }
                return false;
            },

            getCurrentFileData()
            {
                if (this.focusedFile && this.fileData[this.focusedFile]) {
                    return this.fileData[this.focusedFile].file;
                }
                else {
                    return null;
                }
            },

            getProcessing()
            {
                if (this.appConfig.processing[this.appConfig.service]) {
                    return this.appConfig.processing[this.appConfig.service];
                }
                else {
                    return this.appConfig.processing[''];
                }
            },

            getProtectedFilesBadge()
            {
                let res = '';
                if (this.protectedFiles.length) {
                    if (this.protectedFiles.length == 1) {
                        res += `<span class="badge badge-warning text-dark">${IMatchTranslate.getStr('T_STARTPAGE_FILECOUNT_PROTECTED_ONE')}</span>`;
                    }
                    else {
                        let s = IMatchTranslate.getStr('T_STARTPAGE_FILECOUNT_PROTECTED_MULTI');
                        s = s.replace('{0}',this.protectedFiles.length.toLocaleString());
                        res += `<span class="badge badge-warning text-dark">${s}</span>`;
                    }
                }
                return res;
            },

            getStartPageFileCounts()
            {
                let res = IMatchTranslate.getStr('T_STARTPAGE_FILECOUNT');
                res = res.replace('{0}',this.selectedFiles.length);
                res = res.replace('{1}',this.filesWithKeywords.length);
                res = res.replace('{f1}',this.selectedFiles.length == 1 ? IMatchTranslate.getStr('T_FILE') : IMatchTranslate.getStr('T_FILES'));
                res = res.replace('{f2}',this.filesWithKeywords.length == 1 ? IMatchTranslate.getStr('T_FILE') : IMatchTranslate.getStr('T_FILES'));
                if (this.protectedFiles.length) {
                    res += ' ' + this.getProtectedFilesBadge;
                }
                return res;
            },

            getResultKeywords()
            {
                if (this.appConfig.sortAlphabetically) {
                    return this.getCurrentFileResults.keywords.sort((a,b) => {
                        return a.keyword.localeCompare(b.keyword);
                    });
                }
                else {
                    return this.getCurrentFileResults.keywords.sort((a,b) => {
                        return b.score - a.score;
                    });
                }
            },

            getRunResult()
            {
                let res = IMatchTranslate.getStr('T_AUTOPAGE_COMPLETED').replace('{0}',this.runState.filesProcessed).replace('{1}',this.runState.keywordsAdded);
                res = res.replace('{f}',this.runState.filesProcessed == 1 ? IMatchTranslate.getStr('T_FILE') : IMatchTranslate.getStr('T_FILES'));
                
                if (this.getProcessing.skipFilesWithKeywords) {
                    // Did we skip files?
                    let skipped = this.selectedFiles.length - this.runState.filesProcessed;
                    if (skipped > 0) {
                        let s = IMatchTranslate.getStr('T_AUTOPAGE_COMPLETED_SKIPPED').replace('{0}',skipped);
                        res += '<br/><small>' + s.replace('{f}',skipped == 1 ? IMatchTranslate.getStr('T_FILE') : IMatchTranslate.getStr('T_FILES')) + '</small>';
                    }
                }
                if (this.runState.filesSkipped) {
                    let s = IMatchTranslate.getStr('T_AUTOPAGE_COMPLETED_SKIPPED_INVALID').replace('{0}',this.runState.filesSkipped);
                    res += '<br/><small>' + s.replace('{f}',this.runState.filesSkipped == 1 ? IMatchTranslate.getStr('T_FILE') : IMatchTranslate.getStr('T_FILES')) + '</small>';
                }

                res += '<br/>'+IMatchTranslate.getStr('T_AUTOPAGE_RUNTIME').replace('{0}',Math.ceil(this.runState.totalTime / 1000));
                
                return res;
            },

            getCategoryLink()
            {
                return IMatchTranslate.getStr('T_AUTOPAGE_CATLINK').replace('{0}',this.getProcessing.category);
            },
            
            getRunResultCanceled()
            {
                let res = IMatchTranslate.getStr('T_AUTOPAGE_CANCELED').replace('{0}',this.runState.filesProcessed).replace('{1}',this.files.length);
                res = res.replace('{f}',this.files.length == 1 ? IMatchTranslate.getStr('T_FILE') : IMatchTranslate.getStr('T_FILES'));
                return res;
            }

        },

        methods: {

            _setBusy(busy)
            {
                this.busy = busy;
            },

            getServiceName(service)
            {
                if (service.name[IMatchTranslate.getLoadedLangId()] !== undefined) {
                    return service.name[IMatchTranslate.getLoadedLangId()];
                }
                else {
                    return service.name['en'];
                }
            },

            openLink(url)
            {
                IMatch.openLinkInDefaultBrowser(url);
            },

            openCategory()
            {
                IMatch.gotoCategory(this.categoryId);
                window.setTimeout(() => {
                    IMatch.modalClose();
                },50);
            },

            closeDialog()
            {
                window.setTimeout(() => {
                    IMatch.modalClose();
                },50);
            },

            openHelp(target)
            {
                if (target) {
                    IMatch.openIMatchWebHelp(`app-auto-tagger.htm#${target}`);
                }
                else {
                    IMatch.openIMatchWebHelp('app-auto-tagger.htm');
                }
            },

            disableKeyboardHints()
            {
                this.appConfig.showKeyboardHints = false;
                this.delayedSaveConfig();
            },

            // Load the app configuration
            loadConfig() {
                IMWS.dataGet({
                    name: DATA_CONFIG
                }).then(response => {
                    var data = JSON.parse(response.value);
                    if (data.version == DATA_CONFIG_VERSION) {
                        this.appConfig = data;
                    }
                    else {
                        this.appConfig = data;
                        if (this.appConfig.showKeyboardHints === undefined) {
                            this.appConfig.showKeyboardHints = true;
                        }
                    }
                }).always(() => {

                    this.engine = new IMATEngine(this.appConfig.processing,this.simulate);

                    $.getJSON('services.json').done(data => {
                        this.serviceInfo = data;
                        this._initService();

                        if (!this.serviceValid()) {
                            this.stage = 'serviceInvalid';
                        }

                        this._loadFiles();

                        this._checkCategory();

                        //this.closeStartPage('configure');                        
                    });
                });
            },

            // Store the app configuration
            delayedSaveConfig: IMatch.debounce(function(id) {
                vm.saveConfig();
            },100),

            saveConfig() {
                this.appConfig.version = DATA_CONFIG_VERSION;
                IMWS.dataSet({
                    name: DATA_CONFIG,
                    value: JSON.stringify(this.appConfig)
                });
            },

            // Creaate a service instance for the service selected by the user.
            _initService()
            {
                if (this.serviceInstance) {
                    this.serviceInstance = null;
                }

                if (this.appConfig.service != '') {

                    // Validate the service info.
                    const sinfo = this.serviceInfo[this.appConfig.service];
                    if (sinfo === undefined) {
                        // Service no longer exists. Reset back to no service selected.
                        this.appConfig.service = '';
                    }
                    else {
                        const ssettings = this.appConfig.serviceSettings[this.appConfig.service];
                        if (ssettings === undefined) {
                            this.appConfig.service = '';
                        }
                        else {
                            // Check if the selected model still exists.
                            if (sinfo.models[ssettings.model] === undefined) {
                                // Fall back to the first model.
                                this.appConfig.serviceSettings[this.appConfig.service].model = Object.keys(sinfo.models)[0];
                            }
                        }
                    }

                    if (this.appConfig.service != '') {
                        switch (this.appConfig.service) {
                        case 'GOOGLEVISION':
                                this.serviceInstance = new IMATServiceGoogle(this.serviceInfo,this.appConfig.serviceSettings,this.engine,this.appConfig.service);
                            break;
                        case 'MSAZVISION':
                        case 'MSAZVISION-FREE_TIER':
                                this.serviceInstance = new IMATServiceMicrosoft(this.serviceInfo,this.appConfig.serviceSettings,this.engine,this.appConfig.service);
                            break;
                        case 'CLARIFAI':
                            this.serviceInstance = new IMATServiceClarifai(this.serviceInfo,this.appConfig.serviceSettings,this.engine,this.appConfig.service);
                            break;
                        case 'IMAGGA':
                            this.serviceInstance = new IMATServiceImagga(this.serviceInfo,this.appConfig.serviceSettings,this.engine,this.appConfig.service);
                            break;
                        case 'IMATCH':
                            this.serviceInstance = new IMATServiceIMatch(this.serviceInfo,this.appConfig.serviceSettings,this.engine,this.appConfig.service);
                        }

                        // Make sure the stats exist (in case of a service not used before)
                        if (!this.appConfig.serviceStats[this.appConfig.service]) {
                            this.appConfig.serviceStats[this.appConfig.service] = {
                                apiCalls: 0,
                                sinceDate: moment().unix()
                            }
                        }
                        let s = this.appConfig.serviceStats[this.appConfig.service];
                        IMatch.addLogFileEntry(`AutoTagger: Service: ${this.appConfig.service} with ${s.apiCalls} calls since ${s.sinceDate}`);
                        IMatch.addTelemetryEntry({
                            type: 'metric',
                            name: 'app.autotag.service',
                            text: this.appConfig.service
                        });
                    }
                }

                // Make sure we have the proper data for the current service.
                this._updateServiceStats(0);
            },

            serviceValid() {
                return  (this.appConfig.service != '')
                        && (this.appConfig.service.apiKey != '')
                        &&  (this.appConfig.serviceSettings[this.appConfig.service].features.length);
            },

            _cleanString(s)
            {
                return s.replace(/[<,>,&]/g,'_');
            },

            _showError(content,addInfo)
            {
                let m = new IMatchModals(4);

                if (typeof(addInfo) == 'object') {
                    let s = JSON.stringify(addInfo,null,2);
                    if (s == '{}') {
                        s = JSON.stringify('{' + addInfo + '}',null,2);
                    }
                    addInfo = this._cleanString(s);
                }

                IMatch.addLogFileEntry(`${LOG_NAME} Error: '${content}' '${addInfo}'.`);

                m.showErrorBox({
                    headline: IMatchTranslate.getStr('T_ERROR_CAPTION'),
                    content: this._cleanString(content),
                    icon: 'far fa-frown fa-3x',
                    additionalInfo: addInfo,
                });
            },

            _updateServiceStats(apiCalls)
            {
                if (this.appConfig.service != '') {
                    let stats = this.appConfig.serviceStats[this.appConfig.service];
                    stats.apiCalls += apiCalls;
                    let doSave = apiCalls > 0;
                    // If we have no "since" date yet, we use the current date and time
                    if (!stats.sinceDate) {
                        stats.sinceDate = moment().unix();
                        doSave = true;
                    }
                    if (doSave) {
                        this.delayedSaveConfig();                
                    }
                }
            },

            resetServiceStats()
            {
                let stats = this.appConfig.serviceStats[this.appConfig.service];
                stats.apiCalls = 0;
                stats.sinceDate = moment().unix();
                this.delayedSaveConfig();
            },

            closeStartPage(mode)
            {
                if (mode == 'auto') {
                    this.stage = 'auto';
                    this.processFiles();
                }
                else if (mode == 'manual') {
                    this.stage = 'manual';
                    this._loadFiles();
                }
                else if (mode == 'configure') {
                    this.openConfigure();
                }
            },

            // Load all selected files
            _loadFiles()
            {
                this.focusedFile = 0;
                this.files = [];
                this.fileData = [];
                this._setBusy(true);

                if (this.stage == 'manual') {
                    this.fileWindowLayout.layout().type = 'strip';
                    this.fileWindowLayout.layout().elements = this.appConfig.thumbSize;
                    this.fileWindowLayout.layout().squareThumbs = this.appConfig.squareThumbs;
                    this.fileWindowLayout.layout().showProtectedFiles = this.showProtectedFiles;
                    this.fileWindowLayout.layout().dynamic = false;
                    this.fileWindowLayout.layout().hierarchyLevels = 0;
                    this.fileWindowLayout.layout().hideStackedFiles = false;
                    this.fileWindowLayout.layout().directOnly = true;

                    // File name in the footer
                    this.fileWindowLayout.layout().footer = {
                        visible: false,
                        contentFunction(file, layout) {
                            let res = '';
                            res += `<small>${file.name}</small>`;
                            return res;
                        }
                    }

                    this.fileWindowLayout.finalize();
                }

                // We load the selection in the current file window
                let options = {
                    source: 'files'
                }

                // Get the files to process from the active file window.
                // We skip all files which are protected unless the display of
                // protected files is enabled.
                IMWS.get('v1/files',{
                    idlist: IMatch.idlist.fileWindowSelection,
                    fields: 'id,protected',
                    tagkeywords: 'hierarchicalkeywords',
                    taglat: 'gpslatitude',
                    taglng: 'gpslongitude'
                }).then(response => {
                    if (response.files) {
                        this.files = [];
                        this.filesWithKeywords = [];
                        this.selectedFiles = [];
                        this.protectedFiles = [];
                        for (let f of response.files) {
                            this.selectedFiles.push(f);
                            let hasKeywords = f.keywords && f.keywords.length;
                            if (this.showProtectedFiles || !f.protected) {
                                // If enabled, skip files which already have keywords.
                                if ((this.stage == 'manual') || (!this.getProcessing.skipFilesWithKeywords || !hasKeywords)) {
                                    this.files.push(f.id);
                                }
                            }
                            else {
                                this.protectedFiles.push(f.id);
                            }
                            if (hasKeywords) {
                                this.filesWithKeywords.push(f.id);
                            }
                        }

                        IMatch.addLogFileEntry(`${LOG_NAME} ${this.files.length} files in scope.`);

                        // Get the focused file in the file window.
                        // We use that as the initial focus for the strip window.
                        IMWS.get('v1/files',{
                            idlist: IMatch.idlist.fileWindowFocusedFile,
                            fields: 'id'
                        }).then(response => {
                            if (response.files) {
                                this.focusedFile = response.files[0];
                            }

                            if (this.focusedFile == 0) {
                                if (this.files.length) {
                                    this.focusedFile = this.files[0];
                                }
                            }
                            if (this.focusedFile != 0) {
                                if (this.stage == 'manual') {
                                    this.delayedLoadFileData(this.focusedFile,true);
                                }
                            }

                            // We let the loader load the same files we use in this class.
                            if ((this.stage == 'manual') && this.files.length) {
                                options.id = this.files;
                                this.fileWindowLayout.layout().focusFile = this.focusedFile;
                                this.fileWindowLoader.load(options, this.fileWindowLayout.layout()).then(() => {
                                });
                            }
                            
                            this._setBusy(false);

                        }, error => {
                            this._showError(IMatchTranslate.getStr('T_ERROR_IMWS'),error);
                            this._setBusy(false);
                        });
                    }
                },error => {
                    this._showError(IMatchTranslate.getStr('T_ERROR_IMWS'),error);
                    this._setBusy(false);
                });
            },

            // Load the data for the focusd file
            delayedLoadFileData: IMatch.debounce(function(id,force) {
                vm._loadFileData(id,force);
            },100),

            _loadFileData(id,force)
            {
                if (id != 0) {

                    // Already loaded?
                    if (!force && (this.fileData[id] && this.fileData[id].file)) {
                        return;
                    }

                    let fd = {}

                    IMWS.get('v1/files',{
                        id: id,
                        fields: 'id,filename,name,protected'
                    }).then(response => {
                        if (response.files) {
                            // Get the keywords for this file, we need the colors too.
                            IMWS.get('v1/files/keywords',{
                                id: id,
                                colorcodes: true
                            }).then(kwr => {
                                if (kwr.files) {
                                    fd = response.files[0];
                                    fd.keywords = kwr.files[0].keywords;
                                    for (let k of fd.keywords) {
                                        k.selected =  true;
                                    }
                                    // Make this reactive
                                    if (this.fileData[id] == undefined) {
                                        Vue.set(this.fileData, id, {file: fd});
                                    }
                                    else {
                                        Vue.set(this.fileData[id], 'file', fd);
                                    }
                                }
                            });
                        }
                    });
                }
            },

            _processFile(id,batchMode)
            {
                return new Promise((resolve, reject) => {
                    this.serviceInstance.processFile(id,batchMode).then(result => {
                        resolve(result);
                        if (!batchMode) {
                            IMatch.addTelemetryEntry({
                                type: 'count',
                                name: 'app.autotag.files',
                                number: 1
                            });
                        }
                    }).catch(error => {
                        reject(error);
                    });
                });
            },

            // If a category is configured for the current processing,
            // we ensure that it exists and fetch the id.
            _checkCategory()
            {
                this.categoryId = 0;
                let self = this;

                function createCategory()
                {
                    IMWS.post('v1/categories/create',{
                        parentid: 0,
                        name: self.getProcessing.category
                    }).then(response => {
                        if (response.result == 'ok') {
                            self.categoryId = response.id[0];
                        }
                        else {
                            self._showError(IMatchTranslate.getStr('T_ERROR_CREATE_CATEGORY'),response.resultMessage);
                        }
                    }, error => {
                        self._showError(IMatchTranslate.getStr('T_ERROR_CREATE_CATEGORY'),error);
                    });
                }

                if (this.getProcessing.category.length) {
                    IMWS.get('v1/categories',{
                        path: this.getProcessing.category,
                        fields: 'id'
                    }).then(response => {
                        if (response.categories.length) {
                            this.categoryId = response.categories[0].id;
                        }
                        else {
                            createCategory();
                        }
                    },error => {
                        // Categroy does not exist, create it
                        createCategory();
                    });
                }
            },

            onSelectionChanged()
            {
            },

            onFocusChanged()
            {
                this.focusedFile = this.fileWindowLayout.layout().focusedFile;
                this.delayedLoadFileData(this.focusedFile,false);

                this.firstLoadedFile = false;

                // Do we update the File Window?
                if (this.appConfig.updateFileWindowFocus) {
                    IMatch.focusFile({
                        id: this.focusedFile,
                        select: true,
                        resetselection: false
                    });
                }
            },

            // Lookup the keywords for the current file.
            onBtnLookup()
            {
                if (this.focusedFile == 0) return;

                // We set the app to busy if this is an external service
                // or it is the first request for the IMatch AI (which is slow due to model loading).
                if (this.firstLookup || !this.usesIMatchService()) {
                    this._setBusy(true);
                    this.firstLookup = false;
                }

                if (this.fileData[this.focusedFile]) {
                    delete this.fileData[this.focusedFile].result;
                }

                this.$nextTick(() => {
                    this._processFile(this.focusedFile,false).then(result => {
                        this._updateServiceStats(this.serviceInstance.stats().apiCalls);

                        // Need to use Vue.set to make this change (adding a new element on the root level) to fileData reactive.
                        // See: https://vuejs.org/v2/guide/reactivity.html
                        if (this.fileData[this.focusedFile] == undefined) {
                            Vue.set(this.fileData, this.focusedFile, {});
                        }
                        Vue.set(this.fileData[this.focusedFile],'result',result);

                        // Add the "selected" element to each keyword
                        for (let k of this.fileData[this.focusedFile].result.keywords) {
                            // If the option to skip unmapped keyword is on, we 
                            // don't select these. Else we select all.
                            if (this.appConfig.processing[''].skipUnmapped) {
                                k.selected = k.mapped.length;
                            }
                            else {
                                k.selected = true;
                            }
                        }
                    })
                    .catch(error => {
                        if (error === 404) {
                            // The file has no thumbnail.
                            this._showError(IMatchTranslate.getStr('T_ERROR_NO_IMAGE'));
                        }
                        else {
                            this._updateServiceStats(this.serviceInstance.stats().apiCalls);
                            this._showError(IMatchTranslate.getStr('T_ERROR_SERVICE'),error);
                        }
                    })
                    .finally(() => {
                        this._setBusy(false);
                    });
                });
            },

            onBtnConfigure()
            {
                this.openConfigure('');
            },

            usesIMatchService()
            {
                return this.appConfig.service === 'IMATCH';
            },

            // Display the configuration modal
            openConfigure(editMappingFor)
            {
                let cfg = new IMATConfigure(this.appConfig,this.serviceInfo,editMappingFor);
                cfg.run().then(newConfig => {
                    // OK
                    this.appConfig = newConfig;
                    this.firstLookup = true;                    
                    this._initService();
                    this.engine.setProcessing(this.appConfig.processing);
                    this.fileWindowLayout.layout().elements = this.appConfig.thumbSize;
                    this.fileWindowLayout.layout().squareThumbs = this.appConfig.squareThumbs;
                    this.delayedSaveConfig();

                    for (let f in this.fileData) {
                        Vue.delete(this.fileData[f],'result');
                    }

                    this.firstLoadedFile = true;

                    this._loadFiles();

                    if (this.stage == 'serviceInvalid') {
                        if (this.serviceValid()) {
                            this.stage = 'startPage';
                        }
                    }
                }).catch(error => {
                    // Canceled
                });
            },

            onBtnSelect(select) 
            {
                for (let k of this.fileData[this.focusedFile].result.keywords) {
                    k.selected = select;
                }
                this.$forceUpdate();
            },

            onAddMapping(k)
            {
                let m = this.appConfig.processing[''].mapping.maps[k.keyword];
                if (m == undefined) {
                    // New mapping for this keyword
                    this.appConfig.processing[''].mapping.maps[k.keyword] = [];
                }
                this.openConfigure(k.keyword);
            },

            onAddIgnore(k)
            {
                // We ignore based on the original!
                let i = this.appConfig.processing[''].ignore.indexOf(k.original);
                if (i < 0) {
                    this.appConfig.processing[''].ignore.push(k.original);
                    this.delayedSaveConfig();
                    for (let f in this.fileData) {
                        let fd = this.fileData[f];
                        if (fd.result && fd.result.keywords) {
                            let remaining = [];
                            for (const ki of fd.result.keywords) {
                                let i = ki.keyword.indexOf(k.keyword);
                                if (i < 0) {
                                    remaining.push(ki)
                                }
                            }
                            fd.result.keywords = remaining;
                        }
                    }
                }
            },

            // Add a new user-defined keyword
            onAddKeyword()
            {
                let kw = this.newKeyword.trim();
                if (kw.length) {
                    let i = this.fileData[this.focusedFile].result.keywords.indexOf(kw);
                    if (i < 0) {
                        this.fileData[this.focusedFile].result.keywords.push({
                            original: '',
                            keyword: kw,
                            mapped: [],
                            selected: true
                        });
                    }
                    this.newKeyword = '';
                }
                return false;
            },

            onBtnApply()
            {
                if (this.getCurrentFileResults) {
                    this._setBusy(true);
                    let fileData = {};
                    fileData[this.focusedFile] = this.fileData[this.focusedFile].result;
                    this.engine.updateDatabase(fileData,true).then(() => {
                        this.delayedLoadFileData(this.focusedFile,true);
                    }).catch(error => {
                        IMatch.addLogFileEntry(`${LOG_NAME} Error: ${JSON.stringify(error,null,2)}`);
                        this._showError(IMatchTranslate.getStr('T_ERROR_IMWS_APPLYKE'),error)
                    }).finally(() => {
                        this._setBusy(false);
                    });
                }
            },

            onBtnMove(next)
            {
                this.$refs.fileWindow.moveFocus(next);
            },

            onBtnRunCancel()
            {
                this.runState.canceled = true;
                this.setRunning(false,true);
            },

            setRunning(running,canceled)
            {
                this.runState.running = running;
                if (this.runState.running) {
                    this.runState.canceled = false;
                    this.runState.percentage = 0;
                    this.runState.completed = false;                    
                    this.runState.totalTime = 0;
                }
                else {
                    this.runState.completed = true;
                }
            },

            updateTelementry()
            {
                IMatch.addTelemetryEntry({
                    type: 'count',
                    name: 'app.autotag.files',
                    number: this.runState.filesProcessed
                });
            },

            processFiles()
            {
                this.setRunning(true);
                var tStart = performance.now();

                IMatch.addLogFileEntry(`AutoTagger: Processing ${this.files.length} files.`);
                
                let fileIndex = 0;
                let self = this;
                self.runState.completed = false;
                self.runState.error = false;
                self.runState.filesProcessed = 0;
                self.runState.filesSkipped = 0;
                self.runState.keywordsAdded = 0;
                self.fileData = {};

                // How many requests to make per minute max.
                let waitTime = 0;
                if (this.serviceInstance.requetsPerMinute() > 0) {
                    waitTime = Math.ceil(1000 / (this.serviceInstance.requetsPerMinute() / 60));
                }
                
                // How many files can be processed in one batch?
                // let batchSize = this.serviceInstance.batchSize();
                let cache = {};
                let cacheLen = 0;
                const CACHE_SIZE = 50;

                async function flush()
                {
                    if (cacheLen > 0) {
                        try {
                            await self.engine.updateDatabase(cache,false);
                        }
                        catch(error) {
                            IMatch.addLogFileEntry(`${LOG_NAME} Error: ${JSON.stringify(error,null,2)}`);
                            self._showError(IMatchTranslate.getStr('T_ERROR_IMWS_APPLYKE'),error)
                            self.runState.canceled = true;
                        };
                    }
                    cache = {}
                    cacheLen = 0;
                }

                async function nextFile()
                {
                    if (self.runState.canceled) {
                        // Aborted
                        flush();
                        self._loadFiles();
                        self.setRunning(false,true);
                        self.runState.totalTime = performance.now() - tStart;
                        self.updateTelementry();
                    }
                    else if (fileIndex >= self.files.length) {
                        // Done
                        flush();
                        self._loadFiles();
                        self.setRunning(false,false);
                        self.runState.totalTime = performance.now() - tStart;
                        self.updateTelementry();
                    }
                    else {
                        let id = self.files[fileIndex];

                        // Fetch the keywords
                        let result;
                        try {
                            result = await self._processFile(id,true);
                        }
                        catch(error) {
                            if (error === 404) {
                                // File has no thumbnail. Log and proceed with the next file.
                                self.runState.filesSkipped++;
                            }
                            else {
                                self._updateServiceStats(self.serviceInstance.stats().apiCalls);
                                self.runState.error = true;
                                self._showError(IMatchTranslate.getStr('T_ERROR_SERVICE'),error)
                                self.setRunning(false,false);
                                IMatch.addLogFileEntry('AutoTagger: Error processing files: ' + error);
                                self._loadFiles();
                            }
                        }

                        if (result) {
                            self._updateServiceStats(self.serviceInstance.stats().apiCalls);
                            Vue.set(self.fileData,id,{
                                result: result
                            });

                            self.runState.keywordsAdded += result.keywords.length;
                            cache[id] = self.fileData[id].result;
                            cacheLen++;
                            if (cacheLen >= CACHE_SIZE) {
                                await flush();
                            }

                            self.runState.filesProcessed++;
                        }
                        
                        // Update stats etc. and continue with the next file.
                        fileIndex++;
                        self.runState.percentage = Math.ceil((fileIndex * 100) / self.files.length);

                        if (waitTime > 0) {
                            window.setTimeout(() => {
                                nextFile();
                            },waitTime);
                        }
                        else {
                            window.setTimeout(() => {
                                nextFile();
                            },5);
                        } 
                    }
                }
                
                // Start everything
                nextFile();
            }

        }
    });
});
